#include <Arduino.h>
#include <WiFi.h>
#include <ESPmDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <driver/adc.h>

#if defined(ESP8266)
#include <ESP8266WiFi.h>
#else
#include <WiFi.h>
#endif

//needed for library
#include <DNSServer.h>
#if defined(ESP8266)
#include <ESP8266WebServer.h>
#else
#include <WebServer.h>
#endif
#include <WiFiManager.h>

#include "ESP32_cnc.h"
#include "channels.h"
#include "staging_buffer.h"

channels_t serial_channels, udp_channels;
staging_buffer_t staging_buffer;

WiFiUDP Udp;
unsigned int localUdpPort = 4210;  // local port to listen on

#define UDP_MAX_SIZE 512
uint8_t UdpInPacket[UDP_MAX_SIZE];  // buffer for incoming packets
uint8_t UdpOutPacket[UDP_MAX_SIZE];  // buffer for outgoing packets
int UdpBufferSize = UDP_MAX_SIZE;
int UDP_count;


/* change ssid and password according to yours WiFi*/
//const char* ssid     = "";
//const char* password = "";

/*
 * This is the IP address of your PC
 * [Wins: use ipconfig command, Linux: use ip addr command]
*/

//const char* host = "192.168.1.73";
//const int port = 8088;



/*// create a hardware timer
hw_timer_t * timer = NULL;


void IRAM_ATTR onTimer(void)
{
  state = !state;
  digitalWrite(led, state);
}
*/


const int led = 2;
const int power_enable = 19;


volatile uint32_t state = LOW;

unsigned long previousMillis, interval;
int low_voltage_count, low_voltage_tresh;

void serial_write(uint8_t b)
{
  Serial.write(b);
}

void udp_write(uint8_t b)
{
  staging_buffer.sendByte(b);
}

void handle_full_UDP_buffer(uint8_t* buf)
{
  // send back a reply, to the IP address and port we got the packet from
  Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
  Udp.write((uint8_t*)buf, staging_buffer.get_size());
  Udp.endPacket();
  staging_buffer.empty();
}


void process_serial_packet(char channel, uint32_t value, channels_t& obj)
{
  int i, a, ctrl;
  int16_t ival;

  if (channel == 'S') {          // Set new speed and acceleration
    i = (value >> 24) & 0xFF;    // MSB is the motor number
    if (i >= NUM_MOTORS) return; // There are only 2 motors
    ival = value & 0xFFFF;       // LSW has the speed in two's complement
    a = (value >> 16) & 0xFF;    // 0 -> the speed is immediatly set
    if (a == 0) {                // Other value -> the new acceleration
      set_motor_req_speed(i, ival);
      set_motor_speed(i, ival);
    } else {
      set_motor_accel(i, a);
      set_motor_req_speed(i, ival);
    }

  } else if (channel == 'P')  {  // Set new position
    i = (value >> 24) & 0xFF;    // MSB is the motor number
    if (i >= NUM_MOTORS) return; // There are only 2 motors
    ival = value & 0xFFFF;       // LSW has the position in two's complement
    ctrl = (value >> 16) & 0xFF;
    if (ctrl == 0) {
      set_motors_ctrl_state(i, false);
      set_motor_pos(i, ival);
    } else {
      set_motors_ctrl_state(i, true);
      set_motors_pos_ref(i, ival);
    }

  } else if (channel == 'I')  {
    interval = value;

  } else if (channel == 'T')  {
    low_voltage_tresh = value;

  } else if (channel == 'G')  { // Ping
    obj.send(channel, value + 1);
    Serial.println(value + 1);
  }
}



void setup()
{
  digitalWrite(power_enable, 0);
  pinMode(power_enable, OUTPUT);

  pinMode(led, OUTPUT);
  low_voltage_count = 0;
  low_voltage_tresh = 2560; // equivalent to 10.0 V

  Serial.begin(115200);
  serial_channels.init(process_serial_packet, serial_write);

  udp_channels.init(process_serial_packet, udp_write);
  staging_buffer.init(UdpOutPacket, handle_full_UDP_buffer, UdpBufferSize);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;
  //reset saved settings
  //wifiManager.resetSettings();

  //set custom ip for portal
  //wifiManager.setAPStaticIPConfig(IPAddress(10,0,1,1), IPAddress(10,0,1,1), IPAddress(255,255,255,0));

  //fetches ssid and pass from eeprom and tries to connect
  //if it does not connect it starts an access point with the specified name
  //here  "AutoConnectAP"
  //and goes into a blocking loop awaiting configuration
  //wifiManager.autoConnect("AutoConnectAP");
  //or use this for auto generated name ESP + ChipID
  wifiManager.autoConnect();

  /*
  Serial.print("Connecting to ");
  Serial.println(ssid);
  // connect to your WiFi
  WiFi.begin(ssid, password);
  // wait until ESP32 connect to WiFi
  while (WiFi.status() != WL_CONNECTED) {
      delay(200);
      Serial.print(".");
      state = !state;
      digitalWrite(led, state);
  }*/
  Serial.println("");
  Serial.println("WiFi connected with IP address: ");
  Serial.println(WiFi.localIP());

  // Port defaults to 3232
  // ArduinoOTA.setPort(3232);

  // Hostname defaults to esp3232-[MAC]
  // ArduinoOTA.setHostname("myesp32");

  // No authentication by default
  // ArduinoOTA.setPassword("admin");

  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");

  ArduinoOTA
      .onStart([]() {
        String type;
        if (ArduinoOTA.getCommand() == U_FLASH)
          type = "sketch";
        else // U_SPIFFS
          type = "filesystem";

        // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
        Serial.println("Start updating " + type);
      })
      .onEnd([]() {
        Serial.println("\nEnd");
      })
      .onProgress([](unsigned int progress, unsigned int total) {
        Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
      })
      .onError([](ota_error_t error) {
        Serial.printf("Error[%u]: ", error);
        if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
        else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
        else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
        else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
        else if (error == OTA_END_ERROR) Serial.println("End Failed");
      });

  ArduinoOTA.begin();

  /*// Use 1st timer of 4
  // 1 tick take 1/(80MHZ/80) = 1us so we set divider 80 and count up
  timer = timerBegin(0, 80, true);

  // Attach onTimer function to our timer
  timerAttachInterrupt(timer, &onTimer, true);

  // Set alarm to call onTimer function every second 1 tick is 1us  => 1 second is 1000000us
  // Repeat the alarm (third parameter)
  timerAlarmWrite(timer, 100000, true);

  // Start an alarm
  timerAlarmEnable(timer);

  Serial.println("start timer");*/

  ESP32CNC_init();
  adc1_config_width(ADC_WIDTH_12Bit);
  adc1_config_channel_atten(ADC1_CHANNEL_6, ADC_ATTEN_2_5db);

  interval = 50000UL;

  UDP_count = 0;
  Udp.begin(localUdpPort);
  Serial.printf("Now listening at IP %s, UDP port %d\n", WiFi.localIP().toString().c_str(), localUdpPort);


  Serial.println("Init done");

}




void loop()
{
  int adc_voltage;
  int i, c, serialByte;

  // Process Serial Port messages
  while (Serial.available() > 0) {
    serialByte = Serial.read();
    serial_channels.StateMachine(serialByte);
    //Serial.write(serialByte);
  }

  int packetSize = Udp.parsePacket();
  if (packetSize) {
    // receive incoming UDP packets
    Serial.printf("Received %d bytes from %s, port %d\n", packetSize, Udp.remoteIP().toString().c_str(), Udp.remotePort());
    int len = Udp.read(UdpInPacket, UdpBufferSize - 1);
    if (len > 0) {
      UdpInPacket[len] = 0;
    }
    UDP_count++;
    Serial.printf("UDP packet contents (as string): %s\n", UdpInPacket);

    for (i = 0; i < len; i++) {
      udp_channels.StateMachine(UdpInPacket[i]);
    }

    if (!staging_buffer.is_empty()) {
      // send back a reply, to the IP address and port we got the packet from
      Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
      Udp.write(staging_buffer.get_buffer(), staging_buffer.get_size());
      Udp.endPacket();
      staging_buffer.empty();
    }

    //Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
    //Udp.beginPacket(IPAddress(192,168,0,101), 55056);
    //Udp.write(replyPacekt);
    //Udp.endPacket();
  }


  // Read battery voltage and switch off if low ->  < 10.0 V
  adc_voltage = adc1_get_raw(ADC1_CHANNEL_6);  // adc1_get_raw(esp32_get_channel(pin));
  if (adc_voltage < low_voltage_tresh) {
    low_voltage_count++;
  } else {
    low_voltage_count = 0;
  }

  if (low_voltage_count > 10000)
    digitalWrite(power_enable, 1);  // switch off
  else
    digitalWrite(power_enable, 0);  // switch on


  // Switch off drivers if all speeds are zero
  /*
  c = 1;
  for (i = 0; i < NUM_MOTORS; i++) {
    if (get_motor_speed(i) || get_motor_req_speed(i)) {
      c = 0;
      break;  // if any speed is not zero, the motors must be enabled
    }
  }
  digitalWrite(EN, c); //Low to enable
  */

  unsigned long currentMillis = micros();

  if(currentMillis - previousMillis > interval) {
    previousMillis = currentMillis;

    // Update position controllers (only updated if motors_pos_ctrl = TRUE)
    update_motors_pos_ctrl();

    // At least, one UDP packet must has been received so that we know the remote IP and port
    if (UDP_count) {
      staging_buffer.send('V', adc_voltage);
      staging_buffer.send('M', get_motor_speed(0));
      staging_buffer.send('N', get_motor_speed(1));
      staging_buffer.send('P', get_motor_pos(0));
      staging_buffer.send('Q', get_motor_pos(1));
      set_motor_pos(1,0);
    }

    // If there are messages to be sent
    if (staging_buffer.get_size()) {
      // Append a CR/LF
      staging_buffer.sendByte(0x0D);
      staging_buffer.sendByte(0x0A);
      // Send the UDP packet
      handle_full_UDP_buffer(staging_buffer.get_buffer());
    }

    //Serial.print(adc_voltage);
    //Serial.println();
  }

  ArduinoOTA.handle();
}
